@@ -2,7 +2,7 @@
 #
 # SNMP Snapshot Data Recorder
 #
-# Written by Ilya Etingof <ilya@glas.net>, 2010-2013
+# Written by Ilya Etingof <ilya@glas.net>, 2010-2015
 #
 
 import getopt
@@ -24,12 +24,15 @@ try:
 except ImportError:
     unix = None
 from pysnmp.entity.rfc3413 import cmdgen
-from pysnmp import debug
+from pysnmp.smi import rfc1902, view, compiler
+from pyasn1 import debug as pyasn1_debug
+from pysnmp import debug as pysnmp_debug
 from snmpsim.record import snmprec
 from snmpsim import confdir, error, log
 
 # Defaults
 getBulkFlag = False
+continueOnErrors = 0
 getBulkRepetitions = 25
 snmpVersion = 1
 snmpCommunity = 'public'
@@ -38,13 +41,18 @@ v3AuthKey = None
 v3PrivKey = None
 v3AuthProto = 'NONE'
 v3PrivProto = 'NONE'
+v3ContextEngineId = None
 v3Context = ''
 agentUDPv4Address = (None, 161)  # obsolete
 agentUDPv4Endpoint = None
 agentUDPv6Endpoint = None
 agentUNIXEndpoint = None
+timeout = 300  # 1/100 sec
+retryCount = 3
 startOID = univ.ObjectIdentifier('1.3.6')
 stopOID = None
+mibSources = []
+defaultMibSources = [ 'http://mibs.snmplabs.com/asn1/@mib@' ]
 outputFile = sys.stdout
 if hasattr(outputFile, 'buffer'):
     outputFile = outputFile.buffer
@@ -71,7 +79,8 @@ helpMessage = """\
 Usage: %s [--help]
     [--version]
     [--debug=<%s>]
-    [--logging-method=<stdout|stderr|syslog|file>[:args>]]
+    [--debug-asn1=<%s>]
+    [--logging-method=<%s[:args>]>]
     [--protocol-version=<1|2c|3>]
     [--community=<string>]
     [--v3-user=<username>]
@@ -79,36 +88,47 @@ Usage: %s [--help]
     [--v3-auth-proto=<%s>]
     [--v3-priv-key=<key>]
     [--v3-priv-proto=<%s>]
-    [--context=<string>]
+    [--v3-context-engine-id=<[0x]string>]
+    [--v3-context-name=<[0x]string>]
     [--use-getbulk]
     [--getbulk-repetitions=<number>]
     [--agent-udpv4-endpoint=<X.X.X.X:NNNNN>]
     [--agent-udpv6-endpoint=<[X:X:..X]:NNNNN>]
     [--agent-unix-endpoint=</path/to/named/pipe>]
-    [--start-oid=<OID>] [--stop-oid=<OID>]
+    [--timeout=<seconds>] [--retries=<count>]
+    [--mib-source=<url>]
+    [--start-object=<MIB-NAME::[symbol-name]|OID>]
+    [--stop-object=<MIB-NAME::[symbol-name]|OID>]
     [--output-file=<filename>]
     [--variation-modules-dir=<dir>]
     [--variation-module=<module>]
-    [--variation-module-options=<args>]""" % (
+    [--variation-module-options=<args>]
+    [--continue-on-errors=<max-sustained-errors>]""" % (
         sys.argv[0],
-        '|'.join([ x for x in debug.flagMap.keys() if x != 'mibview' ]),
+        '|'.join([ x for x in pysnmp_debug.flagMap.keys() if x != 'mibview' ]),
+        '|'.join([ x for x in pyasn1_debug.flagMap.keys() ]),
+        '|'.join(log.gMap.keys()),
         '|'.join([ x for x in authProtocols if x != 'NONE' ]),
         '|'.join([ x for x in privProtocols if x != 'NONE' ])
     )
 
-log.setLogger('snmprec', 'stderr')
-
 try:
     opts, params = getopt.getopt(sys.argv[1:], 'hv', [
-        'help', 'version', 'debug=', 'logging-method=', 'quiet',
+        'help', 'version', 'debug=', 'debug-asn1=', 'logging-method=', 'quiet',
         'v1', 'v2c', 'v3', 'protocol-version=', 'community=', 
         'v3-user=', 'v3-auth-key=', 'v3-priv-key=', 'v3-auth-proto=',
-        'v3-priv-proto=', 'context=', 'use-getbulk', 'getbulk-repetitions=',
-        'agent-address=', 'agent-port=', 'agent-udpv4-endpoint=',
-        'agent-udpv6-endpoint=', 'agent-unix-endpoint=', 
-        'start-oid=', 'stop-oid=', 'output-file=',
+        'v3-priv-proto=',
+        'context-engine-id=', 'v3-context-engine-id=',
+        'context=', 'v3-context-name=',
+        'use-getbulk', 'getbulk-repetitions=', 'agent-address=', 'agent-port=',
+        'agent-udpv4-endpoint=', 'agent-udpv6-endpoint=',
+        'agent-unix-endpoint=', 'timeout=', 'retries=',
+        'start-oid=', 'stop-oid=',
+        'mib-source=',
+        'start-object=', 'stop-object=',
+        'output-file=',
         'variation-modules-dir=', 'variation-module=',
-        'variation-module-options='
+        'variation-module-options=', 'continue-on-errors='
     ] )
 except Exception:
     sys.stderr.write('ERROR: %s\r\n%s\r\n' % (sys.exc_info()[1], helpMessage))
@@ -131,25 +151,27 @@ Documentation:
 """ % helpMessage)
         sys.exit(-1)
     if opt[0] == '-v' or opt[0] == '--version':
-        import snmpsim, pysnmp, pyasn1
+        import snmpsim, pysnmp, pysmi, pyasn1
         sys.stderr.write("""\
 SNMP Simulator version %s, written by Ilya Etingof <ilya@glas.net>
-Using foundation libraries: pysnmp %s, pyasn1 %s.
+Using foundation libraries: pysmi %s, pysnmp %s, pyasn1 %s.
 Python interpreter: %s
 Software documentation and support at http://snmpsim.sf.net
 %s
-""" % (snmpsim.__version__, hasattr(pysnmp, '__version__') and pysnmp.__version__ or 'unknown', hasattr(pyasn1, '__version__') and pyasn1.__version__ or 'unknown', sys.version, helpMessage))
+""" % (snmpsim.__version__, hasattr(pysmi, '__version__') and pysmi.__version__ or 'unknown', hasattr(pysnmp, '__version__') and pysnmp.__version__ or 'unknown', hasattr(pyasn1, '__version__') and pyasn1.__version__ or 'unknown', sys.version, helpMessage))
         sys.exit(-1)
-    elif opt[0] == '--debug':
-        debug.setLogger(debug.Debug(opt[1]))
+    elif opt[0] in ('--debug', '--debug-snmp'):
+        pysnmp_debug.setLogger(pysnmp_debug.Debug(*opt[1].split(','), **dict(loggerName='snmprec.pysnmp')))
+    elif opt[0] == '--debug-asn1':
+        pyasn1_debug.setLogger(pyasn1_debug.Debug(*opt[1].split(','), **dict(loggerName='snmprec.pyasn1')))
     elif opt[0] == '--logging-method':
         try:
-            log.setLogger('snmprec', *opt[1].split(':'))
+            log.setLogger('snmprec', *opt[1].split(':'), **dict(force=True))
         except error.SnmpsimError:
             sys.stderr.write('%s\r\n%s\r\n' % (sys.exc_info()[1], helpMessage))
             sys.exit(-1)
     elif opt[0] == '--quiet':
-        log.setLogger('snmprec', 'null') 
+        log.setLogger('snmprec', 'null', force=True) 
     elif opt[0] == '--v1':
         snmpVersion = 0
     elif opt[0] == '--v2c':
@@ -184,8 +206,16 @@ Software documentation and support at http://snmpsim.sf.net
         if v3PrivProto not in privProtocols:
             sys.stderr.write('ERROR: bad v3 privacy protocol %s\r\n%s\r\n' % (v3PrivProto, helpMessage))
             sys.exit(-1)
-    elif opt[0] == '--context':
-        v3Context = opt[1]
+    elif opt[0] in ('--v3-context-engine-id', '--context-engine-id'):
+        if opt[1][:2] == '0x':
+            v3ContextEngineId = univ.OctetString(hexValue=opt[1][2:])
+        else:
+            v3ContextEngineId = univ.OctetString(opt[1])
+    elif opt[0] in ('--v3-context-name', '--context'):
+        if opt[1][:2] == '0x':
+            v3Context = univ.OctetString(hexValue=opt[1][2:])
+        else:
+            v3Context = univ.OctetString(opt[1])
     elif opt[0] == '--use-getbulk':
         getBulkFlag = True
     elif opt[0] == '--getbulk-repetitions':
@@ -231,10 +261,30 @@ Software documentation and support at http://snmpsim.sf.net
             sys.stderr.write('This system does not support UNIX domain sockets\r\n')
             sys.exit(-1)
         agentUNIXEndpoint = opt[1]
+    elif opt[0] == '--timeout':
+        try:
+            timeout = float(opt[1])*100
+        except:
+            sys.stderr.write('ERROR: improper --timeout value %s\r\n%s\r\n' % (opt[1], helpMessage))
+            sys.exit(-1)
+    elif opt[0] == '--retries':
+        try:
+            retryCount = int(opt[1])
+        except:
+            sys.stderr.write('ERROR: improper --retries value %s\r\n%s\r\n' % (opt[1], helpMessage))
+            sys.exit(-1)
+    # obsolete begin
     elif opt[0] == '--start-oid':
         startOID = univ.ObjectIdentifier(opt[1])
     elif opt[0] == '--stop-oid':
         stopOID = univ.ObjectIdentifier(opt[1])
+    # obsolete end
+    elif opt[0] == '--mib-source':
+        mibSources.append(opt[1])
+    elif opt[0] == '--start-object':
+        startOID = rfc1902.ObjectIdentity(*opt[1].split('::'))
+    elif opt[0] == '--stop-object':
+        stopOID = rfc1902.ObjectIdentity(*opt[1].split('::'), **dict(last=True))
     elif opt[0] == '--output-file':
         outputFile = open(opt[1], 'wb')
     elif opt[0] == '--variation-modules-dir':
@@ -243,6 +293,12 @@ Software documentation and support at http://snmpsim.sf.net
         variationModuleName = opt[1]
     elif opt[0] == '--variation-module-options':
         variationModuleOptions = opt[1]
+    elif opt[0] == '--continue-on-errors':
+        try:
+            continueOnErrors = int(opt[1])
+        except:
+            sys.stderr.write('ERROR: improper --continue-on-errors retries count %s\r\n%s\r\n' % (opt[1], helpMessage))
+            sys.exit(-1)
  
 # Catch missing params
 
@@ -274,7 +330,12 @@ if snmpVersion == 3:
         if v3PrivKey is None:
             sys.stderr.write('ERROR: --v3-priv-key is missing\r\n%s\r\n' % helpMessage)
             sys.exit(-1)
- 
+else:
+    v3ContextEngineId = None
+    v3ContextName = ''
+
+log.setLogger('snmprec', 'stderr')
+
 if getBulkFlag and not snmpVersion:
     log.msg('WARNING: will be using GETNEXT with SNMPv1!')
     getBulkFlag = False
@@ -335,7 +396,7 @@ if snmpVersion == 3:
         authProtocols[v3AuthProto], v3AuthKey,
         privProtocols[v3PrivProto], v3PrivKey
         )
-    log.msg('SNMP version 3, Context name: %s, SecurityName: %s, SecurityLevel: %s, Authentication key/protocol: %s/%s, Encryption (privacy) key/protocol: %s/%s' % (v3Context == '' and '\'\'' or v3Context, v3User, secLevel, v3AuthKey is None and '<NONE>' or v3AuthKey, v3AuthProto, v3PrivKey is None and '<NONE>' or v3PrivKey, v3PrivProto))
+    log.msg('SNMP version 3, Context EngineID: %s Context name: %s, SecurityName: %s, SecurityLevel: %s, Authentication key/protocol: %s/%s, Encryption (privacy) key/protocol: %s/%s' % (v3ContextEngineId and v3ContextEngineId.prettyPrint() or '<default>', v3Context and v3Context.prettyPrint() or '<default>', v3User, secLevel, v3AuthKey is None and '<NONE>' or v3AuthKey, v3AuthProto, v3PrivKey is None and '<NONE>' or v3PrivKey, v3PrivProto))
 else:
     v3User = 'agt'
     secLevel = 'noAuthNoPriv'
@@ -353,7 +414,8 @@ if agentUDPv6Endpoint:
         udp6.Udp6SocketTransport().openClientMode()
     )
     config.addTargetAddr(
-        snmpEngine, 'tgt', udp6.domainName, agentUDPv6Endpoint, 'pms'
+        snmpEngine, 'tgt', udp6.domainName, agentUDPv6Endpoint, 'pms',
+        timeout, retryCount
     )
     log.msg('Querying UDP/IPv6 agent at [%s]:%s' % agentUDPv6Endpoint)
 elif agentUNIXEndpoint:
@@ -363,7 +425,8 @@ elif agentUNIXEndpoint:
         unix.UnixSocketTransport().openClientMode()
     )
     config.addTargetAddr(
-        snmpEngine, 'tgt', unix.domainName, agentUNIXEndpoint, 'pms'
+        snmpEngine, 'tgt', unix.domainName, agentUNIXEndpoint, 'pms',
+        timeout, retryCount
     )
     log.msg('Querying UNIX named pipe agent at %s' % agentUNIXEndpoint)
 elif agentUDPv4Endpoint:
@@ -373,10 +436,25 @@ elif agentUDPv4Endpoint:
         udp.UdpSocketTransport().openClientMode()
     )
     config.addTargetAddr(
-        snmpEngine, 'tgt', udp.domainName, agentUDPv4Endpoint, 'pms'
+        snmpEngine, 'tgt', udp.domainName, agentUDPv4Endpoint, 'pms',
+        timeout, retryCount
     )
     log.msg('Querying UDP/IPv4 agent at %s:%s' % agentUDPv4Endpoint)
 
+log.msg('Agent response timeout: %.2f secs, retries: %s' % (timeout/100, retryCount))
+
+if isinstance(startOID, rfc1902.ObjectIdentity) or \
+        isinstance(stopOID, rfc1902.ObjectIdentity):
+    compiler.addMibCompiler(
+        snmpEngine.getMibBuilder(),
+        sources=mibSources or defaultMibSources
+    )
+    mibViewController = view.MibViewController(snmpEngine.getMibBuilder())
+    if isinstance(startOID, rfc1902.ObjectIdentity):
+        startOID.resolveWithMib(mibViewController)
+    if isinstance(stopOID, rfc1902.ObjectIdentity):
+        stopOID.resolveWithMib(mibViewController)
+
 # Variation module initialization
 
 if variationModule:
@@ -427,17 +505,67 @@ dataFileHandler = SnmprecRecord()
 
 # SNMP worker
 
-def cbFun(sendRequestHandle, errorIndication, errorStatus, errorIndex,
-          varBindTable, cbCtx):
-    if errorIndication and errorIndication != 'oidNotIncreasing':
+def cbFun(snmpEngine, sendRequestHandle, errorIndication,
+          errorStatus, errorIndex, varBindTable, cbCtx):
+    if errorIndication and not cbCtx['retries']:
+        cbCtx['errors'] += 1
         log.msg('SNMP Engine error: %s' % errorIndication)
         return
     # SNMPv1 response may contain noSuchName error *and* SNMPv2c exception,
     # so we ignore noSuchName error here
-    if errorStatus and errorStatus != 2:
-        log.msg('Remote SNMP error: %s' % errorStatus.prettyPrint())
+    if errorStatus and errorStatus != 2 or errorIndication:
+        log.msg('Remote SNMP error %s' % (errorIndication or errorStatus.prettyPrint()))
+        if cbCtx['retries']:
+            try:
+                nextOID = varBindTable[-1][0][0]
+            except IndexError:
+                nextOID = cbCtx['lastOID']
+            else:
+                log.msg('Failed OID: %s' % nextOID)
+            # fuzzy logic of walking a broken OID
+            if len(nextOID) < 4:
+                pass
+            elif (continueOnErrors-cbCtx['retries'])*10/continueOnErrors > 5:
+                nextOID = nextOID[:-2] + (nextOID[-2]+1,)
+            elif nextOID[-1]:
+                nextOID = nextOID[:-1] + (nextOID[-1]+1,)
+            else:
+                nextOID = nextOID[:-2] + (nextOID[-2]+1, 0)
+
+            cbCtx['retries'] -= 1
+            cbCtx['lastOID'] = nextOID
+
+            log.msg('Retrying with OID %s (%s retries left)...' % (nextOID, cbCtx['retries']))
+            
+            # initiate another SNMP walk iteration
+            if getBulkFlag:
+                cmdGen.sendVarBinds(
+                    snmpEngine, 
+                    'tgt', 
+                    v3ContextEngineId, v3Context,
+                    0, getBulkRepetitions,
+                    [ (nextOID, None) ], 
+                    cbFun, cbCtx
+                )
+            else:
+                cmdGen.sendVarBinds(
+                    snmpEngine,
+                   'tgt',
+                    v3ContextEngineId, v3Context,
+                    [ (nextOID, None) ],
+                    cbFun, cbCtx
+                )
+
+        cbCtx['errors'] += 1
+
         return
 
+    if continueOnErrors != cbCtx['retries']:
+        cbCtx['retries'] += 1
+
+    if varBindTable and varBindTable[-1] and varBindTable[-1][0]:
+        cbCtx['lastOID'] = varBindTable[-1][0][0]
+
     stopFlag = False
 
     # Walk var-binds
@@ -480,16 +608,24 @@ def cbFun(sendRequestHandle, errorIndication, errorStatus, errorIndex,
  
                 # initiate another SNMP walk iteration
                 if getBulkFlag:
-                    cmdGen.sendReq(
-                        snmpEngine, 'tgt', 0, getBulkRepetitions, 
-                        ((startOID, None),), cbFun, cbCtx,
-                        contextName=v3Context
+                    cmdGen.sendVarBinds(
+                        snmpEngine, 
+                        'tgt', 
+                        v3ContextEngineId, v3Context,
+                        0, getBulkRepetitions,
+                        [ (startOID, None) ], 
+                        cbFun, cbCtx
                     )
                 else:
-                    cmdGen.sendReq(
-                        snmpEngine, 'tgt', ((startOID, None),), cbFun, cbCtx,
-                        contextName=v3Context
+                    cmdGen.sendVarBinds(
+                        snmpEngine,
+                        'tgt',
+                        v3ContextEngineId, v3Context,
+                        [ (startOID, None) ],
+                        cbFun, cbCtx
                     )
+
+                stopFlag = True  # stop current iteration
  
             except error.NoDataNotification:
                 pass
@@ -499,11 +635,11 @@ def cbFun(sendRequestHandle, errorIndication, errorStatus, errorIndex,
             else:
                 outputFile.write(line)
 
-            cbCtx['count'] += 1
-            cbCtx['total'] += 1
+                cbCtx['count'] += 1
+                cbCtx['total'] += 1
 
-            if cbCtx['count'] % 100 == 0:
-                log.msg('OIDs dumped: %s/%s' % (cbCtx['iteration'], cbCtx['count']))
+                if cbCtx['count'] % 100 == 0:
+                    log.msg('OIDs dumped: %s/%s' % (cbCtx['iteration'], cbCtx['count']))
 
     # Next request time
     cbCtx['reqTime'] = time.time()
@@ -514,26 +650,36 @@ def cbFun(sendRequestHandle, errorIndication, errorStatus, errorIndex,
 cbCtx = {
     'total': 0,
     'count': 0,
+    'errors': 0,
     'iteration': 0,
-    'reqTime': time.time()
+    'reqTime': time.time(),
+    'retries': continueOnErrors,
+    'lastOID': startOID
 }
 
 if getBulkFlag:
     cmdGen = cmdgen.BulkCommandGenerator()
 
-    cmdGen.sendReq(
-        snmpEngine, 'tgt', 0, getBulkRepetitions, ((startOID, None),), 
-        cbFun, cbCtx, contextName=v3Context
+    cmdGen.sendVarBinds(
+        snmpEngine,
+        'tgt',
+        v3ContextEngineId, v3Context,
+        0, getBulkRepetitions,
+        [ (startOID, None) ], 
+        cbFun, cbCtx
     )
 else:
     cmdGen = cmdgen.NextCommandGenerator()
 
-    cmdGen.sendReq(
-        snmpEngine, 'tgt', ((startOID, None),),
-        cbFun, cbCtx, contextName=v3Context
+    cmdGen.sendVarBinds(
+        snmpEngine,
+        'tgt',
+        v3ContextEngineId, v3Context,
+        [ (startOID, None) ],
+        cbFun, cbCtx
     )
 
-log.msg('Sending initial %s request....' % (getBulkFlag and 'GETBULK' or 'GETNEXT'))
+log.msg('Sending initial %s request for %s (stop at %s)....' % (getBulkFlag and 'GETBULK' or 'GETNEXT', startOID, stopOID or '<end-of-mib>'))
 
 t = time.time()
 
@@ -565,7 +711,7 @@ t = time.time() - t
 
 cbCtx['total'] += cbCtx['count']
 
-log.msg('OIDs dumped: %s, elapsed: %.2f sec, rate: %.2f OIDs/sec' % (cbCtx['total'], t, t and cbCtx['count']//t or 0))
+log.msg('OIDs dumped: %s, elapsed: %.2f sec, rate: %.2f OIDs/sec, errors: %d' % (cbCtx['total'], t, t and cbCtx['count']//t or 0, cbCtx['errors']))
 
 if exc_info:
     for line in traceback.format_exception(*exc_info):
