@@ -2,7 +2,7 @@
 #
 # SNMP Agent Simulator
 #
-# Written by Ilya Etingof <ilya@glas.net>, 2010-2013
+# Written by Ilya Etingof <ilya@glas.net>, 2010-2015
 #
 import os
 import stat
@@ -33,16 +33,18 @@ from pysnmp.smi import exval, indices
 from pysnmp.smi.error import MibOperationError
 from pysnmp.proto import rfc1902, rfc1905, api
 from pysnmp import error
-from pysnmp import debug
+from pyasn1 import debug as pyasn1_debug
+from pysnmp import debug as pysnmp_debug
 from snmpsim.error import SnmpsimError, NoDataNotification
 from snmpsim import confdir, log, daemon
 from snmpsim.record import dump, mvc, sap, walk, snmprec
-from snmpsim.record.search.file import searchRecordByOid
+from snmpsim.record.search.file import searchRecordByOid, getRecord
 from snmpsim.record.search.database import RecordIndex
 
 # Settings
 forceIndexBuild = False
 validateData = False
+maxVarBinds = 64
 transportIdOffset= 0
 v2cArch = False
 v3Only = False
@@ -142,9 +144,13 @@ class SnmprecRecord(snmprec.SnmprecRecord):
                       recordContexts ) = context['variationModules'][modName]
                     if context['dataFile'] not in agentContexts:
                         agentContexts[context['dataFile']] = {}
+                    if context['dataFile'] not in recordContexts:
+                        recordContexts[context['dataFile']] = {}
                     variationModule['agentContext'] = agentContexts[context['dataFile']]
+                    recordContexts = recordContexts[context['dataFile']]
                     if oid not in recordContexts:
                         recordContexts[oid] = {}
+
                     variationModule['recordContext'] = recordContexts[oid]
 
                     # invoke variation module
@@ -267,12 +273,12 @@ class DataFile(AbstractLayout):
 
             varsRemaining -= 1
 
-            line = text.readline()  # matched line
+            line, _, _ = getRecord(text)  # matched line
  
             while True:
                 if exactMatch:
                     if context.get('nextFlag') and not subtreeFlag:
-                        _nextLine = text.readline() # next line
+                        _nextLine, _, _ = getRecord(text) # next line
                         if _nextLine:
                             _nextOid, _ = self.__textParser.evaluate(_nextLine, oidOnly=True)
                             try:
@@ -304,7 +310,7 @@ class DataFile(AbstractLayout):
                         # previous line serves a subtree?
                         if _prevOffset >= 0:
                             text.seek(_prevOffset)
-                            _prevLine = text.readline()
+                            _prevLine, _, _ = getRecord(text)
                             _prevOid, _ = self.__textParser.evaluate(
                                 _prevLine, oidOnly=True
                             )
@@ -401,17 +407,30 @@ class MibInstrumController:
             return { 'nextFlag': nextFlag,
                      'setFlag': setFlag }
 
-        acFun, acCtx = acInfo
-        ( snmpEngine, 
+        acFun, snmpEngine = acInfo  # we injected snmpEngine object earlier
+        # this API is first introduced in pysnmp 4.2.6 
+        execCtx = snmpEngine.observer.getExecutionContext(
+                'rfc3412.receiveMessage:request'
+        )
+        ( transportDomain,
+          transportAddress,
           securityModel,
           securityName,
           securityLevel,
           contextName, 
-          pduType ) = acCtx  # this is not [yet] documented
+          pduType ) = ( execCtx['transportDomain'],
+                        execCtx['transportAddress'],
+                        execCtx['securityModel'],
+                        execCtx['securityName'],
+                        execCtx['securityLevel'],
+                        execCtx['contextName'],
+                        execCtx['pdu'].getTagSet() )
 
-        log.msg('SNMP EngineID %s, securityModel %s, securityName %s, securityLevel %s' % (hasattr(snmpEngine, 'snmpEngineID') and snmpEngine.snmpEngineID.prettyPrint() or '<unknown>', securityModel, securityName, securityLevel))
+        log.msg('SNMP EngineID %s, transportDomain %s, transportAddress %s, securityModel %s, securityName %s, securityLevel %s' % (hasattr(snmpEngine, 'snmpEngineID') and snmpEngine.snmpEngineID.prettyPrint() or '<unknown>', transportDomain, transportAddress, securityModel, securityName, securityLevel))
 
         return { 'snmpEngine': snmpEngine,
+                 'transportDomain': transportDomain,
+                 'transportAddress': transportAddress,
                  'securityModel': securityModel,
                  'securityName': securityName,
                  'securityLevel': securityLevel,
@@ -499,51 +518,58 @@ helpMessage = """\
 Usage: %s [--help]
     [--version ]
     [--debug=<%s>]
+    [--debug-asn1=<%s>]
     [--daemonize]
     [--process-user=<uname>] [--process-group=<gname>]
     [--pid-file=<file>]
-    [--logging-method=<stdout|stderr|syslog|file>[:args>]]
+    [--logging-method=<%s[:args>]>]
     [--cache-dir=<dir>]
     [--variation-modules-dir=<dir>]
     [--variation-module-options=<module[=alias][:args]>] 
     [--force-index-rebuild]
     [--validate-data]
+    [--args-from-file=<file>]
+    [--transport-id-offset=<number>]
     [--v2c-arch]
     [--v3-only]
-    [--transport-id-offset=<number>]
-    [--args-from-file=<file>]
     [--v3-engine-id=<hexvalue>]
-    [--data-dir=<dir>]
-    [--agent-udpv4-endpoint=<X.X.X.X:NNNNN>]
-    [--agent-udpv6-endpoint=<[X:X:..X]:NNNNN>]
-    [--agent-unix-endpoint=</path/to/named/pipe>]
+    [--v3-context-engine-id=<hexvalue>]
     [--v3-user=<username>]
     [--v3-auth-key=<key>]
     [--v3-auth-proto=<%s>]
     [--v3-priv-key=<key>]
-    [--v3-priv-proto=<%s>]""" % (
+    [--v3-priv-proto=<%s>]
+    [--data-dir=<dir>]
+    [--max-varbinds=<number>]
+    [--agent-udpv4-endpoint=<X.X.X.X:NNNNN>]
+    [--agent-udpv6-endpoint=<[X:X:..X]:NNNNN>]
+    [--agent-unix-endpoint=</path/to/named/pipe>]""" % (
         sys.argv[0],
-        '|'.join([ x for x in debug.flagMap.keys() if x != 'mibview' ]),
+        '|'.join([ x for x in pysnmp_debug.flagMap.keys() if x != 'mibview' ]),
+        '|'.join([ x for x in pyasn1_debug.flagMap.keys() ]),
+        '|'.join(log.gMap.keys()),
         '|'.join(authProtocols),
         '|'.join(privProtocols)
     )
 
 try:
     opts, params = getopt.getopt(sys.argv[1:], 'hv', [
-        'help', 'version', 'debug=', 'daemonize', 'process-user=',
-        'process-group=', 'pid-file=', 'logging-method=', 'device-dir=',
-        'cache-dir=', 'variation-modules-dir=', 
+        'help', 'version', 'debug=', 'debug-snmp=', 'debug-asn1=', 'daemonize',
+        'process-user=', 'process-group=', 'pid-file=', 'logging-method=',
+        'device-dir=', 'cache-dir=', 'variation-modules-dir=', 
         'force-index-rebuild', 'validate-device-data', 'validate-data',
         'v2c-arch', 'v3-only', 'transport-id-offset=', 
         'variation-module-options=',
         'args-from-file=',
         # this option starts new SNMPv3 engine configuration
-        'data-dir=',
         'v3-engine-id=',
-        'agent-udpv4-endpoint=','agent-udpv6-endpoint=','agent-unix-endpoint=',
+        'v3-context-engine-id=',
         'v3-user=',
         'v3-auth-key=', 'v3-auth-proto=', 
-        'v3-priv-key=', 'v3-priv-proto='
+        'v3-priv-key=', 'v3-priv-proto=',
+        'data-dir=',
+        'max-varbinds=',
+        'agent-udpv4-endpoint=','agent-udpv6-endpoint=','agent-unix-endpoint='
     ])
 except Exception:
     sys.stderr.write('ERROR: %s\r\n%s\r\n' % (sys.exc_info()[1], helpMessage))
@@ -553,8 +579,6 @@ if params:
     sys.stderr.write('ERROR: extra arguments supplied %s\r\n%s\r\n' % (params, helpMessage))
     sys.exit(-1)
 
-log.setLogger('snmpsimd', 'stderr')
-
 v3Args = []
 
 for opt in opts:
@@ -579,8 +603,10 @@ Software documentation and support at http://snmpsim.sf.net
 %s
 """ % (snmpsim.__version__, hasattr(pysnmp, '__version__') and pysnmp.__version__ or 'unknown', hasattr(pyasn1, '__version__') and pyasn1.__version__ or 'unknown', sys.version, helpMessage))
         sys.exit(-1)
-    elif opt[0] == '--debug':
-        debug.setLogger(debug.Debug(opt[1]))
+    elif opt[0] in ('--debug', '--debug-snmp'):
+        pysnmp_debug.setLogger(pysnmp_debug.Debug(*opt[1].split(','), **dict(loggerName='snmpsimd.pysnmp')))
+    elif opt[0] == '--debug-asn1':
+        pyasn1_debug.setLogger(pyasn1_debug.Debug(*opt[1].split(','), **dict(loggerName='snmpsimd.pyasn1')))
     elif opt[0] == '--daemonize':
         foregroundFlag = False
     elif opt[0] == '--process-user':
@@ -591,12 +617,13 @@ Software documentation and support at http://snmpsim.sf.net
         pidFile = opt[1]
     elif opt[0] == '--logging-method':
         try:
-            log.setLogger('snmpsimd', *opt[1].split(':'))
+            log.setLogger('snmpsimd', *opt[1].split(':'), **dict(force=True))
         except SnmpsimError:
             sys.stderr.write('%s\r\n%s\r\n' % (sys.exc_info()[1], helpMessage))
             sys.exit(-1)
     elif opt[0] in ('--device-dir', '--data-dir'):
-        if '--v3-engine-id' in [ x[0] for x in v3Args ]:
+        if [ x for x in v3Args if x[0] in ('--v3-engine-id',
+                                           '--v3-context-engine-id') ]:
             v3Args.append(opt)
         else:
             confdir.data.insert(0, opt[1])
@@ -642,9 +669,19 @@ Software documentation and support at http://snmpsim.sf.net
                     '--agent-unix-endpoints-list'):
         sys.stderr.write('ERROR: use --args-from-file=<file> option to list many endpoints\r\n%s\r\n' % helpMessage)
         sys.exit(-1)
-    elif opt[0] in ('--v3-engine-id', '--v3-user', '--v3-auth-key', 
-                    '--v3-auth-proto', '--v3-priv-key', '--v3-priv-proto'):
+    elif opt[0] in ('--v3-engine-id', '--v3-context-engine-id',
+                    '--v3-user', '--v3-auth-key', '--v3-auth-proto',
+                    '--v3-priv-key', '--v3-priv-proto'):
         v3Args.append(opt)
+    elif opt[0] == '--max-varbinds':
+        try:
+            if '--v3-engine-id' in [ x[0] for x in v3Args ]:
+                v3Args.append((opt[0], max(1, int(opt[1]))))
+            else:
+                maxVarBinds = max(1, int(opt[1]))
+        except:
+            sys.stderr.write('ERROR: %s\r\n%s\r\n' % (sys.exc_info()[1], helpMessage))
+            sys.exit(-1)
     elif opt[0] == '--args-from-file':
         try:
             v3Args.extend(
@@ -686,6 +723,8 @@ if v3Args[:len(v3Args)//2] == v3Args[len(v3Args)//2:]:
 else:
     v3Args = v3Args[len(v3Args)//2:]
 
+log.setLogger('snmpsimd', 'stderr')
+
 try:
     daemon.dropPrivileges(procUser, procGroup)
 except:
@@ -740,7 +779,7 @@ for variationModulesDir in confdir.variation:
                 log.msg('Variation module "%s" execution failure: %s' %  (mod, sys.exc_info()[1]))
                 sys.exit(-1)
             else:
-                # moduleContext, agentContext, recordContext
+                # moduleContext, agentContexts, recordContexts
                 variationModules[alias] = ctx, {}, {}
 
     log.msg('A total of %s modules found in %s' % (len(variationModules), variationModulesDir))
@@ -785,7 +824,7 @@ def configureManagedObjects(dataDirs, dataIndexInstrumController,
         if not os.path.exists(dataDir):
             log.msg('Directory "%s" does not exist' % dataDir)
             continue
-        log.msg('%s' % ('='*66,))
+        log.msg.incIdent()
         for fullPath, textParser, communityName in getDataFiles(dataDir):
             if communityName in _dataFiles:
                 log.msg('WARNING: ignoring duplicate Community/ContextName "%s" for data file %s (%s already loaded)' % (communityName, fullPath, _dataFiles[communityName]))
@@ -814,20 +853,24 @@ def configureManagedObjects(dataDirs, dataIndexInstrumController,
                 agentName = contextName = md5(univ.OctetString(communityName).asOctets()).hexdigest()
 
                 if not v3Only:
+                    # snmpCommunityTable::snmpCommunityIndex can't be > 32
                     config.addV1System(
                         snmpEngine, agentName, communityName, contextName=contextName
                     )
 
                 snmpContext.registerContextName(contextName, mibInstrum)
+                
+                if len(communityName) <= 32:
+                    snmpContext.registerContextName(communityName, mibInstrum)
                          
                 dataIndexInstrumController.addDataFile(
                     fullPath, communityName, contextName
                 )
                          
-                log.msg('SNMPv3 context name: %s' % (contextName,))
-                
-            log.msg('%s' % ('-+' * 33,))
-                
+                log.msg('SNMPv3 Context Name: %s%s' % (contextName, len(communityName) <= 32 and ' or %s' % communityName or ''))
+
+        log.msg.decIdent()
+
     del _mibInstrums
     del _dataFiles
 
@@ -836,7 +879,7 @@ if v2cArch:
         N = min(int(nonRepeaters), len(reqVarBinds))
         M = int(maxRepetitions)
         R = max(len(reqVarBinds)-N, 0)
-        if R: M = min(M, 100/R)  # maxVarBinds
+        if R: M = min(M, maxVarBinds/R)
 
         if N:
             rspVarBinds = readNextVars(reqVarBinds[:N])
@@ -938,19 +981,35 @@ if v2cArch:
 
 else: # v3arch
     def probeHashContext(self, snmpEngine, stateReference, contextName):
-        transportDomain, transportAddress = snmpEngine.msgAndPduDsp.getTransportInfo(stateReference)
+        # this API is first introduced in pysnmp 4.2.6 
+        execCtx = snmpEngine.observer.getExecutionContext(
+                                          'rfc3412.receiveMessage:request'
+                                      )
+        ( transportDomain,
+          transportAddress ) = ( execCtx['transportDomain'],
+                                 execCtx['transportAddress'] )
 
-        for candidate in probeContext(transportDomain, transportAddress, contextName):
-            probedContextName = md5(candidate).hexdigest()
+        for candidate in probeContext(transportDomain, transportAddress,
+                                      contextName):
+            if len(candidate) > 32:
+                probedContextName = md5(candidate).hexdigest()
+            else:
+                probedContextName = candidate
             try:
                 mibInstrum = self.snmpContext.getMibInstrum(probedContextName)
             except error.PySnmpError:
                 pass
             else:
                 log.msg('Using %s selected by candidate %s; transport ID %s, source address %s, context name "%s"' % (mibInstrum, candidate, univ.ObjectIdentifier(transportDomain), transportAddress[0], probedContextName))
-                return probedContextName
+                contextName = probedContextName
+                break
+        else:
+            mibInstrum = self.snmpContext.getMibInstrum(contextName)
+            log.msg('Using %s selected by contextName "%s", transport ID %s, source address %s' % (mibInstrum, contextName, univ.ObjectIdentifier(transportDomain), transportAddress[0]))
 
-        log.msg('Using %s selected by contextName "%s", transport ID %s, source address %s' % (self.snmpContext.getMibInstrum(contextName), contextName, univ.ObjectIdentifier(transportDomain), transportAddress[0]))
+        if not isinstance(mibInstrum, (MibInstrumController, DataIndexInstrumController)):
+            log.msg('Warning: LCD access denied (contextName does not match any data file)')
+            raise NoDataNotification()
 
         return contextName
 
@@ -961,8 +1020,10 @@ else: # v3arch
             try:
                 cmdrsp.GetCommandResponder.handleMgmtOperation(
                     self, snmpEngine, stateReference, 
-                    probeHashContext(self, snmpEngine, stateReference, contextName),
-                    PDU, acInfo
+                    probeHashContext(
+                        self, snmpEngine, stateReference, contextName
+                    ),
+                    PDU, (None, snmpEngine) # custom acInfo
                 )
             except NoDataNotification:
                 self.releaseStateInformation(stateReference)
@@ -974,8 +1035,10 @@ else: # v3arch
             try:
                 cmdrsp.SetCommandResponder.handleMgmtOperation(
                     self, snmpEngine, stateReference, 
-                    probeHashContext(self, snmpEngine, stateReference, contextName),
-                    PDU, acInfo
+                    probeHashContext(
+                        self, snmpEngine, stateReference, contextName
+                    ),
+                    PDU, (None, snmpEngine) # custom acInfo
                 )
             except NoDataNotification:
                 self.releaseStateInformation(stateReference)
@@ -987,8 +1050,10 @@ else: # v3arch
             try:
                 cmdrsp.NextCommandResponder.handleMgmtOperation(
                     self, snmpEngine, stateReference, 
-                    probeHashContext(self, snmpEngine, stateReference, contextName),
-                    PDU, acInfo
+                    probeHashContext(
+                        self, snmpEngine, stateReference, contextName
+                    ),
+                    PDU, (None, snmpEngine) # custom acInfo
                 )
             except NoDataNotification:
                 self.releaseStateInformation(stateReference)
@@ -1000,8 +1065,10 @@ else: # v3arch
             try:
                 cmdrsp.BulkCommandResponder.handleMgmtOperation(
                     self, snmpEngine, stateReference, 
-                    probeHashContext(self, snmpEngine, stateReference, contextName),
-                    PDU, acInfo
+                    probeHashContext(
+                        self, snmpEngine, stateReference, contextName
+                    ),
+                    PDU, (None, snmpEngine) # custom acInfo
                 )
             except NoDataNotification:
                 self.releaseStateInformation(stateReference)
@@ -1039,6 +1106,8 @@ if v2cArch:
         log.msg('ERROR: agent endpoint address(es) not specified for SNMP engine ID %s' % v3EngineId)
         sys.exit(-1)
 
+    log.msg('Maximum number of variable bindings in SNMP response: %s' % maxVarBinds)
+
     # Configure socket server
    
     transportIndex = transportIdOffset
@@ -1080,6 +1149,21 @@ else:  # v3 mode
         v3Args.insert(0, ('--v3-engine-id', 'auto'))
     v3Args.append(('end-of-options', ''))
 
+    def registerTransportDispatcher(snmpEngine, transportDispatcher,
+                                    transportDomain):
+        if hasattr(transportDispatcher, 'registerRoutingCbFun'):
+            snmpEngine.registerTransportDispatcher(
+                transportDispatcher, transportDomain
+            )
+        else:
+            try:
+                snmpEngine.registerTransportDispatcher(
+                    transportDispatcher
+                )
+            except error.PySnmpError:
+                log.msg('WARNING: upgrade pysnmp to 4.2.5 or later get multi-engine ID feature working!')
+                raise
+
     snmpEngine = None
     transportIndex = { 'udpv4': transportIdOffset,
                        'udpv6': transportIdOffset,
@@ -1088,21 +1172,39 @@ else:  # v3 mode
     for opt in v3Args:
         if opt[0] in ('--v3-engine-id', 'end-of-options'):
             if snmpEngine:
-                log.msg('%s' % ('*'*66,))
+                log.msg('--- SNMP Engine configuration') 
 
-                dataIndexInstrumController = DataIndexInstrumController()
+                log.msg('SNMPv3 EngineID: %s' % (hasattr(snmpEngine, 'snmpEngineID') and snmpEngine.snmpEngineID.prettyPrint() or '<unknown>',))
 
-                configureManagedObjects(dataDirs and dataDirs or confdir.data, 
-                                        dataIndexInstrumController,
-                                        snmpEngine,
-                                        snmpContext)
+                if not v3ContextEngineIds:
+                    v3ContextEngineIds.append((None, []))
+
+                log.msg.incIdent()
+
+                log.msg('--- Data directories configuration')
+
+                for v3ContextEngineId, ctxDataDirs in v3ContextEngineIds:
+                    snmpContext = context.SnmpContext(
+                        snmpEngine, v3ContextEngineId
+                    )
+
+                    log.msg('SNMPv3 Context Engine ID: %s' % snmpContext.contextEngineId.prettyPrint())
+
+                    dataIndexInstrumController = DataIndexInstrumController()
+
+                    configureManagedObjects(
+                        ctxDataDirs or dataDirs or confdir.data, 
+                        dataIndexInstrumController,
+                        snmpEngine,
+                        snmpContext
+                    )
 
                 # Configure access to data index
 
                 config.addV1System(snmpEngine, 'index',
                                    'index', contextName='index')
 
-                log.msg('SNMPv3 EngineID: %s' % (hasattr(snmpEngine, 'snmpEngineID') and snmpEngine.snmpEngineID.prettyPrint() or '<unknown>',))
+                log.msg('--- SNMPv3 USM configuration')
 
                 if not v3Users:
                     v3Users = [ 'simulator' ]
@@ -1157,7 +1259,9 @@ else:  # v3 mode
                     'index', dataIndexInstrumController
                 )
 
-                # Configure socket server
+                log.msg('Maximum number of variable bindings in SNMP response: %s' % localMaxVarBinds)
+
+                log.msg('--- Transport configuration')
 
                 if not agentUDPv4Endpoints and \
                         not agentUDPv6Endpoints and \
@@ -1169,8 +1273,8 @@ else:  # v3 mode
                     transportDomain = udp.domainName + \
                             (transportIndex['udpv4'],)
                     transportIndex['udpv4'] += 1
-                    snmpEngine.registerTransportDispatcher(
-                        transportDispatcher, transportDomain
+                    registerTransportDispatcher(
+                        snmpEngine, transportDispatcher, transportDomain
                     )
                     config.addSocketTransport(
                         snmpEngine,
@@ -1182,8 +1286,8 @@ else:  # v3 mode
                     transportDomain = udp6.domainName + \
                             (transportIndex['udpv6'],)
                     transportIndex['udpv6'] += 1
-                    snmpEngine.registerTransportDispatcher(
-                        transportDispatcher, transportDomain
+                    registerTransportDispatcher(
+                        snmpEngine, transportDispatcher, transportDomain
                     )
                     config.addSocketTransport(
                         snmpEngine,
@@ -1195,8 +1299,8 @@ else:  # v3 mode
                     transportDomain = unix.domainName + \
                             (transportIndex['unix'],)
                     transportIndex['unix'] += 1
-                    snmpEngine.registerTransportDispatcher(
-                        transportDispatcher, transportDomain
+                    registerTransportDispatcher(
+                        snmpEngine, transportDispatcher, transportDomain
                     )
                     config.addSocketTransport(
                         snmpEngine,
@@ -1208,14 +1312,18 @@ else:  # v3 mode
                 GetCommandResponder(snmpEngine, snmpContext)
                 SetCommandResponder(snmpEngine, snmpContext)
                 NextCommandResponder(snmpEngine, snmpContext)
-                BulkCommandResponder(snmpEngine, snmpContext)
+                BulkCommandResponder(snmpEngine, snmpContext).maxVarBinds = localMaxVarBinds
+
+                log.msg.decIdent()
 
             if opt[0] == 'end-of-options':
                 break
 
             # Prepare for next engine ID configuration
 
+            v3ContextEngineIds = []
             dataDirs = []
+            localMaxVarBinds = maxVarBinds
             v3Users = []
             v3AuthKeys = {}; v3AuthProtos = {}
             v3PrivKeys = {}; v3PrivProtos = {}
@@ -1238,10 +1346,15 @@ else:  # v3 mode
             
             config.addContext(snmpEngine, '')
 
-            snmpContext = context.SnmpContext(snmpEngine)
-
+        elif opt[0] == '--v3-context-engine-id':
+            v3ContextEngineIds.append((univ.OctetString(hexValue=opt[1]), []))
         elif opt[0] == '--data-dir':
-            dataDirs.append(opt[1])
+            if v3ContextEngineIds:
+                v3ContextEngineIds[-1][1].append(opt[1])
+            else:
+                dataDirs.append(opt[1])
+        elif opt[0] == '--max-varbinds':
+            localMaxVarBinds = opt[1]
         elif opt[0] == '--v3-user':
             v3Users.append(opt[1])
             if None in v3AuthKeys:
