@@ -1,10 +1,12 @@
+from string import digits, ascii_letters
 from pysnmp.proto import rfc1902, rfc1905
-from pyasn1.compat.octets import octs2str, str2octs
+from pyasn1.compat.octets import octs2str, str2octs, octs2ints
 from pyasn1.type import univ
 from snmpsim.grammar.abstract import AbstractGrammar
 from snmpsim import error
 
 class SnmprecGrammar(AbstractGrammar):
+    alnums = set(octs2ints(str2octs(ascii_letters+digits)))
     tagMap = {}
     for t in ( rfc1902.Gauge32,
                rfc1902.Integer32,
@@ -21,7 +23,10 @@ class SnmprecGrammar(AbstractGrammar):
                rfc1905.EndOfMibView ):
         tagMap[str(sum([ x for x in t.tagSet[0] ]))] = t
 
-    def build(self, oid, tag, val): return str2octs('%s|%s|%s\n' % (oid, tag, val))
+    def build(self, oid, tag, val):
+        if oid and tag:
+            return str2octs('%s|%s|%s\n' % (oid, tag, val))
+        raise error.SnmpsimError('empty OID/tag <%s/%s>' % (oid, tag))
 
     def parse(self, line):
         try:
@@ -29,7 +34,9 @@ class SnmprecGrammar(AbstractGrammar):
         except:
             raise error.SnmpsimError('broken record <%s>' % line)
         else:
-            return oid, tag, value
+            if oid and tag:
+                return oid, tag, value
+            raise error.SnmpsimError('broken record <%s>' % line)
 
     # helper functions
 
@@ -44,6 +51,6 @@ class SnmprecGrammar(AbstractGrammar):
                             rfc1902.Opaque.tagSet,
                             rfc1902.IpAddress.tagSet):
             nval = value.asNumbers()
-            if nval and nval[-1] == 32 or \
-                    [ x for x in nval if x < 32 or x > 126 ]:
-                return ''.join([ '%.2x' % x for x in nval ])
+            for x in nval:
+                if x not in self.alnums:
+                    return ''.join([ '%.2x' % x for x in nval ])
