@@ -1,5 +1,7 @@
 from pysnmp.proto import rfc1902
-from pyasn1.compat.octets import octs2str
+from pyasn1.type import univ
+from pyasn1.codec.ber import encoder
+from pyasn1.compat.octets import octs2str, ints2octs
 from snmpsim.grammar import abstract
 from snmpsim import error
 
@@ -36,10 +38,14 @@ class WalkGrammar(abstract.AbstractGrammar):
             return value
 
     def __opaqueFilter(value):
-        return [int(y, 16) for y in value.split(' ')]
+        if value.upper().startswith('FLOAT: '):
+            return encoder.encode(univ.Real(float(value[7:])))
+        else:
+            return [int(y, 16) for y in value.split(' ')]
 
     def __bitsFilter(value):
-        return ''.join(['%.2x' % int(y, 16) for y in value.split(' ')])
+        # rfc1902.Bits does not really initialize from sequences
+        return ints2octs([int(y, 16) for y in value.split(' ')])
 
     def __hexStringFilter(value):
         return [int(y, 16) for y in value.split(' ')]
@@ -76,4 +82,6 @@ class WalkGrammar(abstract.AbstractGrammar):
                 value = ''
             else:
                 tag = 'TimeTicks:'
-        return oid, tag.upper(), self.filterMap.get(tag, lambda x: x)(value.strip())
+        if oid and tag:
+            return oid, tag.upper(), self.filterMap.get(tag.upper(), lambda x: x)(value.strip())
+        raise error.SnmpsimError('broken record <%s>' % line)
