@@ -7,6 +7,7 @@ else:
     import dbm
     whichdb = dbm.whichdb
 from snmpsim import confdir, log, error
+from snmpsim.record.search.file import getRecord
 
 class RecordIndex:
     def __init__(self, textFile, textParser):
@@ -81,14 +82,12 @@ class RecordIndex:
                 else:
                     break
             else:
-                log.msg('Failed to create %s for data file %s' % (self.__dbFile, self.__textFile))
-                raise error.SnmpsimError('Cant create data file %s' % self.__dbFile)
+                raise error.SnmpsimError('Failed to create %s for data file %s: %s' % (self.__dbFile, self.__textFile, sys.exc_info()[1]))
 
             try:
                 text = open(self.__textFile, 'rb')
             except:
-                log.msg('Failed to open data file %s: %s' % (self.__dbFile, sys.exc_info()[0]))
-                raise error.SnmpsimError()
+                raise error.SnmpsimError('Failed to open data file %s: %s' % (self.__dbFile, sys.exc_info()[0]))
 
             log.msg('Building index %s for data file %s (open flags \"%s\")...' % (self.__dbFile, self.__textFile, open_flags))
             sys.stdout.flush()
@@ -97,14 +96,12 @@ class RecordIndex:
             offset = 0
             prevOffset = -1
             while 1:
-                line = text.readline()
+                line, lineNo, offset = getRecord(text, lineNo, offset)
                 if not line:
                     # reference to last OID in data file
                     db['last'] = '%d,%d,%d' % (offset, 0, prevOffset)
                     break
-            
-                lineNo += 1
-
+             
                 try:
                     oid, tag, val = self.__textParser.grammar.parse(line)
                 except Exception:
@@ -114,11 +111,7 @@ class RecordIndex:
                         os.remove(self.__dbFile)
                     except OSError:
                         pass
-                    raise error.SnmpsimError(
-                        'Data error at %s:%d: %s' % (
-                            self.__textFile, lineNo, exc
-                            )
-                        )
+                    raise error.SnmpsimError('Data error at %s:%d: %s' % (self.__textFile, lineNo, exc))
 
                 if validateData:
                     try:
@@ -130,11 +123,7 @@ class RecordIndex:
                             os.remove(self.__dbFile)
                         except OSError:
                             pass
-                        raise error.SnmpsimError(
-                            'OID error at %s:%d: %s' % (
-                                self.__textFile, lineNo, exc
-                                )
-                            )
+                        raise error.SnmpsimError('OID error at %s:%d: %s' % (self.__textFile, lineNo, exc))
                     try:
                         self.__textParser.evaluateValue(
                             oid, tag, val, dataValidation=True
@@ -158,7 +147,7 @@ class RecordIndex:
             text.close()
             db.close()
        
-            log.msg('...%d entries indexed' % (lineNo - 1,))
+            log.msg('...%d entries indexed' % lineNo)
 
         self.__textFileTime = os.stat(self.__textFile)[8]
 
