@@ -1,5 +1,20 @@
 from pyasn1.compat.octets import str2octs
 
+# read lines from text file ignoring #comments and blank lines
+def getRecord(fileObj, lineNo=None, offset=0):
+    line = fileObj.readline()
+    if lineNo is not None and line: lineNo += 1
+    while line:
+        tline = line.strip()
+        # skip comment or blank line
+        if not tline or tline.startswith(str2octs('#')):
+            offset += len(line)
+            line = fileObj.readline()
+            if lineNo is not None and line: lineNo += 1
+        else:
+            break
+    return line, lineNo, offset
+
 # In-place, by-OID binary search
 def searchRecordByOid(oid, fileObj, textParser, eol=str2octs('\n')):
     lo = mid = 0; prev_mid = -1
@@ -19,10 +34,12 @@ def searchRecordByOid(oid, fileObj, textParser, eol=str2octs('\n')):
             break
         if mid >= sz:
             return sz
-        line = fileObj.readline()
+        line, _, skippedOffset = getRecord(fileObj)
+        if not line:
+            return hi
         midval, _ = textParser.evaluate(line, oidOnly=True)
         if midval < oid:
-            lo = mid + len(line)
+            lo = mid + skippedOffset + len(line)
         elif midval > oid:
             hi = mid
         else:
@@ -32,4 +49,3 @@ def searchRecordByOid(oid, fileObj, textParser, eol=str2octs('\n')):
         return lo
     else:
         return hi
-
