@@ -1,24 +1,23 @@
 # SNMP Simulator, http://snmpsim.sourceforge.net
 # Managed value variation module
 # Send SNMP Notification
-import sys
-from pysnmp.entity.rfc3413.oneliner import ntforg
-from pysnmp.proto import rfc1902
+from pysnmp.hlapi.asyncore import *
 from snmpsim.grammar.snmprec import SnmprecGrammar
+from snmpsim.mltsplit import split
 from snmpsim import error, log
 
 def init(**context): pass
 
 typeMap = {
-    's': rfc1902.OctetString,
-    'i': rfc1902.Integer32,
-    'o': rfc1902.ObjectName,
-    'a': rfc1902.IpAddress,
-    'u': rfc1902.Unsigned32,
-    'g': rfc1902.Gauge32,
-    't': rfc1902.TimeTicks,
-    'b': rfc1902.Bits,
-    'I': rfc1902.Counter64
+    's': OctetString,
+    'i': Integer32,
+    'o': ObjectIdentifier,
+    'a': IpAddress,
+    'u': Unsigned32,
+    'g': Gauge32,
+    't': TimeTicks,
+    'b': Bits,
+    'I': Counter64
 }
 
 def _cbFun(sendRequestHandle,
@@ -26,16 +25,30 @@ def _cbFun(sendRequestHandle,
           errorStatus, errorIndex,
           varBinds,
           cbCtx):
-    oid, value, ntfOrg = cbCtx
+    oid, value = cbCtx
     if errorIndication or errorStatus:
         log.msg('notification: for %s=%r failed with errorIndication %s, errorStatus %s' % (oid, value, errorIndication, errorStatus))
 
 def variate(oid, tag, value, **context):
     if 'snmpEngine' in context and context['snmpEngine']:
         snmpEngine = context['snmpEngine']
-        if snmpEngine not in recordContext:
-            recordContext[snmpEngine] = ntforg.AsynNotificationOriginator(snmpEngine)
-        ntfOrg = recordContext[snmpEngine]
+        if snmpEngine not in moduleContext:
+            moduleContext[snmpEngine] = {}
+        if context['transportDomain'] not in moduleContext[snmpEngine]:
+            # register this SNMP Engine to handle our transports'
+            # receiver IDs (which we build by outbound and simulator
+            # transportDomains concatenation)
+            snmpEngine.registerTransportDispatcher(
+                snmpEngine.transportDispatcher,
+                UdpTransportTarget.transportDomain + \
+                    context['transportDomain']
+            )
+            snmpEngine.registerTransportDispatcher(
+                snmpEngine.transportDispatcher,
+                Udp6TransportTarget.transportDomain + \
+                    context['transportDomain']
+            )
+            moduleContext[snmpEngine][context['transportDomain']] = 1
     else:
         raise error.SnmpsimError('variation module not given snmpEngine')
 
@@ -43,7 +56,7 @@ def variate(oid, tag, value, **context):
         return context['origOid'], tag, context['errorStatus']
 
     if 'settings' not in recordContext:
-        recordContext['settings'] = dict([ x.split('=') for x in value.split(',') ])
+        recordContext['settings'] = dict([split(x, '=') for x in split(value, ',')])
         for k,v in ( ('op', 'set'),
                      ('community', 'public'),
                      ('authkey', None),
@@ -61,7 +74,7 @@ def variate(oid, tag, value, **context):
 
         if 'vlist' in recordContext['settings']:
             vlist = {}
-            recordContext['settings']['vlist'] = recordContext['settings']['vlist'].split(':')
+            recordContext['settings']['vlist'] = split(recordContext['settings']['vlist'], ':')
             while recordContext['settings']['vlist']:
                 o,v = recordContext['settings']['vlist'][:2]
                 recordContext['settings']['vlist'] = recordContext['settings']['vlist'][2:]
@@ -99,27 +112,27 @@ def variate(oid, tag, value, **context):
        args['op'] == 'set' and context['setFlag'] or \
        args['op'] in ('any', '*'):
         if args['version'] in ('1', '2c'):
-            authData = ntforg.CommunityData(args['community'], args['version'] == '2c' and 1 or 0)
+            authData = CommunityData(args['community'], mpModel=args['version'] == '2c' and 1 or 0)
         elif args['version'] == '3':
             if args['authproto'] == 'md5':
-                authProtocol = ntforg.usmHMACMD5AuthProtocol
+                authProtocol = usmHMACMD5AuthProtocol
             elif args['authproto'] == 'sha':
-                authProtocol = ntforg.usmHMACSHAAuthProtocol
+                authProtocol = usmHMACSHAAuthProtocol
             elif args['authproto'] == 'none':
-                authProtocol = ntforg.usmNoAuthProtocol
+                authProtocol = usmNoAuthProtocol
             else:
                 log.msg('notification: unknown auth proto %s' % args['authproto'])
                 return context['origOid'], tag, context['errorStatus']
             if args['privproto'] == 'des':
-                privProtocol = ntforg.usmDESPrivProtocol
+                privProtocol = usmDESPrivProtocol
             elif args['privproto'] == 'aes':
-                privProtocol = ntforg.usmAesCfb128Protocol
+                privProtocol = usmAesCfb128Protocol
             elif args['privproto'] == 'none':
-                privProtocol = ntforg.usmNoPrivProtocol
+                privProtocol = usmNoPrivProtocol
             else:
                 log.msg('notification: unknown privacy proto %s' % args['privproto'])
                 return context['origOid'], tag, context['errorStatus']
-            authData = ntforg.UsmUserData(args['user'], args['authkey'], args['privkey'], authProtocol=authProtocol, privProtocol=privProtocol)
+            authData = UsmUserData(args['user'], args['authkey'], args['privkey'], authProtocol=authProtocol, privProtocol=privProtocol)
         else:
             log.msg('notification: unknown SNMP version %s' % args['version'])
             return context['origOid'], tag, context['errorStatus']
@@ -127,47 +140,68 @@ def variate(oid, tag, value, **context):
         if 'host' not in args:
             log.msg('notification: target hostname not configured for OID' % (oid,))
             return context['origOid'], tag, context['errorStatus']
-
+        
         if args['proto'] == 'udp':
-            target = ntforg.UdpTransportTarget((args['host'], int(args['port'])))
+            target = UdpTransportTarget((args['host'], int(args['port'])))
         elif args['proto'] == 'udp6':
-            target = ntforg.Udp6TransportTarget((args['host'], int(args['port'])))
+            target = Udp6TransportTarget((args['host'], int(args['port'])))
         else:
             log.msg('notification: unknown transport %s' % args['proto'])
             return context['origOid'], tag, context['errorStatus']
+       
+        localAddress = None
+
+        if 'bindaddr' in args:
+            localAddress = args['bindaddr']
+        else:
+            if context['transportDomain'][:len(target.transportDomain)] == \
+                        target.transportDomain:
+                localAddress = snmpEngine.transportDispatcher.getTransport(context['transportDomain']).getLocalAddress()[0]
+            else:
+                log.msg('notification: incompatible network transport types used by CommandResponder vs NotificationOriginator')
+                if 'bindaddr' in args:
+                    localAddress = args['bindaddr']
+
+        if localAddress:
+            log.msg('notification: binding to local address %s' % localAddress)
+            target.setLocalAddress((localAddress, 0))
+
+        # this will make target objects different based on their bind address 
+        target.transportDomain = target.transportDomain + \
+                                 context['transportDomain']
 
         varBinds = []
 
         if 'uptime' in args:
             varBinds.append(
-                ( rfc1902.ObjectName('1.3.6.1.2.1.1.3.0'),
-                  rfc1902.TimeTicks(args['uptime']) )
+                ( ObjectIdentifier('1.3.6.1.2.1.1.3.0'),
+                  TimeTicks(args['uptime']) )
             )
 
         if args['version'] == '1':
             if 'agentaddress' in args:
                 varBinds.append(
-                    ( rfc1902.ObjectName('1.3.6.1.6.3.18.1.3.0'),
-                      rfc1902.IpAddress(args['agentaddress']) )
+                    ( ObjectIdentifier('1.3.6.1.6.3.18.1.3.0'),
+                      IpAddress(args['agentaddress']) )
                 )
             if 'enterprise' in args:
                 varBinds.append(
-                    ( rfc1902.ObjectName('1.3.6.1.6.3.1.1.4.3.0'),
-                      rfc1902.ObjectName(args['enterprise']) )
+                    ( ObjectIdentifier('1.3.6.1.6.3.1.1.4.3.0'),
+                      ObjectIdentifier(args['enterprise']) )
                 )
 
         if 'varbinds' in args:
-            vbs = args['varbinds'].split(':') 
+            vbs = split(args['varbinds'], ':') 
             while vbs:
                 varBinds.append(
-                    (rfc1902.ObjectName(vbs[0]), typeMap[vbs[1]](vbs[2]))
+                    (ObjectIdentifier(vbs[0]), typeMap[vbs[1]](vbs[2]))
                 )
                 vbs = vbs[3:]
 
-        ntfOrg.sendNotification(
-            authData, target, args['ntftype'],
-            rfc1902.ObjectName(args['trapoid']),
-            varBinds, cbInfo=(_cbFun, (oid, value, ntfOrg))
+        sendNotification(
+            snmpEngine, authData, target, ContextData(), args['ntftype'],
+            NotificationType(ObjectIdentity(args['trapoid'])).addVarBinds(*varBinds),
+            cbFun=_cbFun, cbCtx=(oid, value)
         )
 
         log.msg('notification: sending Notification to %s with credentials %s' % (authData, target))
