@@ -10,23 +10,28 @@
 # CREATE TABLE <tablename> (oid text, tag text, value text, maxaccess text)
 #
 from snmpsim.grammar.snmprec import SnmprecGrammar
-from snmpsim import error
+from snmpsim.mltsplit import split
+from snmpsim import error, log
 
-dbConn = None
-dbTable = 'snmprec'
+isolationLevels = {
+    '0': 'READ UNCOMMITTED',
+    '1': 'READ COMMITTED',
+    '2': 'REPEATABLE READ',
+    '3': 'SERIALIZABLE'
+}
 
 def init(**context):
-    global dbConn, dbTable
     options = {}
     if context['options']:
         options.update(
-            dict([x.split(':') for x in context['options'].split(',')])
+            dict([split(x, ':') for x in split(context['options'], ',')])
         )
     if 'dbtype' not in options:
         raise error.SnmpsimError('database type not specified')
     db = __import__(
         options['dbtype'], 
-        fromlist=options['dbtype'].split('.')[:-1]
+        globals(), locals(),
+        options['dbtype'].split('.')[:-1]
     )
     if 'dboptions' in options: # legacy
         connectParams = { 'database': options['dboptions'] }
@@ -39,26 +44,57 @@ def init(**context):
                 connectParams[k] = int(connectParams[k])
     if not connectParams:
         raise error.SnmpsimError('database connect parameters not specified')
-    dbConn = db.connect(**connectParams)
-    if 'dbtable' in options:
-        dbTable = options['dbtable']
+    moduleContext['dbConn'] = dbConn = db.connect(**connectParams)
+    moduleContext['dbTable'] = dbTable = options.get('dbtable', 'snmprec')
+    moduleContext['isolationLevel'] = options.get('isolationlevel', '1')
+    if moduleContext['isolationLevel'] not in isolationLevels:
+        raise error.SnmpsimError('unknown SQL transaction isolation level %s' % moduleContext['isolationLevel'])
+
     if 'mode' in context and context['mode'] == 'recording':
         cursor = dbConn.cursor()
+
         try:
-            cursor.execute('select * from %s limit 1' % dbTable)
-        except:
-            cursor.execute('CREATE TABLE %s (oid text, tag text, value text, maxaccess text)' % dbTable)
+            cursor.execute(
+                'select * from information_schema.tables where table_name=\'%s\' limit 1' % dbTable
+            )
+        except: # non-ANSI database
+            try:
+                cursor.execute('select * from %s limit 1' % dbTable)
+            except:
+                createTable = True
+            else:
+                createTable = False
         else:
-            cursor.fetchall()
+            createTable = not cursor.fetchone()
+
+        if createTable:
+            cursor.execute('CREATE TABLE %s (oid text, tag text, value text, maxaccess text)' % dbTable)
+
         cursor.close()
 
 def variate(oid, tag, value, **context):
-    if dbConn is None:
+    if 'dbConn' in moduleContext:
+        dbConn = moduleContext['dbConn']
+    else:
         raise error.SnmpsimError('variation module not initialized')
 
     cursor = dbConn.cursor()
 
-    dbTable, = value.split(',')
+    try:
+        cursor.execute(
+            'set session transaction isolation level %s' % moduleContext['isolationLevel']
+        )
+        cursor.fetchall()
+    except:  # non-MySQL/Postgres
+        pass
+
+    if value:
+        dbTable = value.split(',').pop(0)
+    elif 'dbTable' in moduleContext:
+        dbTable = moduleContext['dbTable']
+    else:
+        log.msg('SQL table not specified for OID %s' % (context['origOid'],))
+        return context['origOid'], tag, context['errorStatus']
 
     origOid = context['origOid']
     sqlOid = '.'.join(['%10s' % x for x in str(origOid).split('.')])
@@ -71,7 +107,7 @@ def variate(oid, tag, value, **context):
             textTag = SnmprecGrammar().getTagByType(context['origValue'])
             textValue = str(context['origValue'])
         cursor.execute(
-            'select maxaccess,tag from %s where oid="%s"' % (dbTable, sqlOid)
+            'select maxaccess,tag from %s where oid=\'%s\' limit 1' % (dbTable, sqlOid)
         )
         resultset = cursor.fetchone()
         if resultset:
@@ -79,11 +115,11 @@ def variate(oid, tag, value, **context):
             if maxaccess != 'read-write':
                 return origOid, tag, context['errorStatus']
             cursor.execute(
-                'update %s set tag="%s",value="%s" where oid="%s"' % (dbTable, textTag, textValue, sqlOid)
+                'update %s set tag=\'%s\',value=\'%s\' where oid=\'%s\'' % (dbTable, textTag, textValue, sqlOid)
             )
         else:
             cursor.execute(
-                'insert into %s values ("%s", "%s", "%s", "read-write")' % (dbTable, sqlOid, textTag, textValue)
+                'insert into %s values (\'%s\', \'%s\', \'%s\', \'read-write\')' % (dbTable, sqlOid, textTag, textValue)
             )
         if context['varsRemaining'] == 0:  # last OID in PDU
             dbConn.commit()
@@ -91,7 +127,7 @@ def variate(oid, tag, value, **context):
         return origOid, textTag, context['origValue']
     else:
         if context['nextFlag']:
-            cursor.execute('select oid from %s where oid>"%s" order by oid limit 1' % (dbTable, sqlOid))
+            cursor.execute('select oid from %s where oid>\'%s\' order by oid limit 1' % (dbTable, sqlOid))
             resultset = cursor.fetchone()
             if resultset:
                 origOid = origOid.clone(
@@ -102,7 +138,7 @@ def variate(oid, tag, value, **context):
                 cursor.close()
                 return origOid, tag, context['errorStatus']
 
-        cursor.execute('select tag, value from %s where oid="%s"' % (dbTable, sqlOid))
+        cursor.execute('select tag, value from %s where oid=\'%s\' limit 1' % (dbTable, sqlOid))
         resultset = cursor.fetchone()
         cursor.close()
 
@@ -112,9 +148,13 @@ def variate(oid, tag, value, **context):
             return origOid, tag, context['errorStatus']
 
 def record(oid, tag, value, **context):
-    if dbConn is None:
+    if 'dbConn' in moduleContext:
+        dbConn = moduleContext['dbConn']
+    else:
         raise error.SnmpsimError('variation module not initialized')
 
+    dbTable = moduleContext['dbTable']
+
     if context['stopFlag']:
         raise error.NoDataNotification()
 
@@ -129,15 +169,15 @@ def record(oid, tag, value, **context):
     cursor = dbConn.cursor()
 
     cursor.execute(
-        'select oid from %s where oid="%s" limit 1' % (dbTable, sqlOid)
+        'select oid from %s where oid=\'%s\' limit 1' % (dbTable, sqlOid)
     )
     if cursor.fetchone():
         cursor.execute(
-            'update %s set tag="%s",value="%s" where oid="%s"' % (dbTable, textTag, textValue, sqlOid)
+            'update %s set tag=\'%s\',value=\'%s\' where oid=\'%s\'' % (dbTable, textTag, textValue, sqlOid)
         )
     else:
         cursor.execute(
-            'insert into %s values ("%s", "%s", "%s", "read-write")' % (dbTable, sqlOid, textTag, textValue)
+            'insert into %s values (\'%s\', \'%s\', \'%s\', \'read-write\')' % (dbTable, sqlOid, textTag, textValue)
         )
     cursor.close()
 
@@ -147,7 +187,8 @@ def record(oid, tag, value, **context):
         raise error.NoDataNotification()
 
 def shutdown(**context):
-    if dbConn is not None:
+    dbConn = moduleContext.get('dbConn')
+    if dbConn:
         if 'mode' in context and context['mode'] == 'recording':
             dbConn.commit()
         dbConn.close()
